with Unchecked_Deallocation;

package body Contenedores.AVLs is
	--------------------------
	-- Operaciones Internas --
	--------------------------	
	procedure Eliminar is new Unchecked_Deallocation(Nodo_AVL, rbol_AVL);

	procedure Rotar_simple_AVL(A: in out Rbol_AVL; Rama: in integer) is
		OtraRama: Integer := 3 - Rama;
		Hijo    : Rbol_AVL := A.Hijos(Rama);
	begin
		-- Se intercambian los nodos
		A.Hijos(Rama)        := Hijo.Hijos(OtraRama);
		Hijo.Hijos(OtraRama) := A;
		A                    := Hijo;

		-- Se ajustan los factores de equilibrio
		Hijo := A.Hijos(OtraRama);
		Hijo.Equilibrio := Hijo.Equilibrio - A.Equilibrio;
		A.Equilibrio    := -Hijo.Equilibrio;
	end Rotar_Simple_AVL;

	Procedure Rotar_Doble_AVL(A: in out Rbol_AVL; Rama: in Integer) is
		OtraRama: Integer := 3 - Rama;
		Hijo    : Rbol_AVL := A.Hijos(Rama);
		OtroHijo: Rbol_AVL;
		Nieto   : Rbol_AVL := Hijo.Hijos(OtraRama);
	begin
		-- Se intercambian los nodos
		A.Hijos(Rama)         := Nieto.Hijos(OtraRama);
		Hijo.Hijos(OtraRama)  := Nieto.Hijos(Rama);
		Nieto.Hijos(OtraRama) := A;
 	    Nieto.Hijos(Rama)     := Hijo;
		A                     := Nieto;

		-- Se ajustan los factores de equilibrio
		Hijo                := A.Hijos(Rama);
		OtroHijo            := A.Hijos(OtraRama);
		Hijo.Equilibrio     := (1 - 2 * (rama - 1) - A.Equilibrio) / 2;
      	otroHijo.Equilibrio := (2 * (rama - 1) - A.Equilibrio - 1) / 2;
		a.Equilibrio        := 0;
	end Rotar_Doble_AVL;

	procedure Insertar(
			A: in out rbol_AVL;
			E: in TElemento;
			aumenta_Altura: in out boolean
			) is
		Rama: Integer;
		Aux : rbol_AVL;
		Factor_Equi, Factor_Equi_Hijo: integer range -1 .. 1;
	begin
		if A = null then
			-- El nuevo valor se inserta como hoja
			A := new Nodo_AVL'(0,E, (null, null));
			Aumenta_Altura := True;
		else
			-- Si E = A.Clave, no se inserta
			if E /= A.Clave then
				-- La insercin continua en el subrbol adecuado
				if E < A.Clave then
					Rama := 1;
					Factor_Equi := 1;
				else
					Rama := 2;
					Factor_Equi := -1;
				end if;

				Insertar(A.Hijos(Rama), E, aumenta_Altura);
				Factor_Equi_Hijo := A.Hijos(Rama).Equilibrio;

				-- Si la altura aumenta hay que comprobar los
				-- factores de equilibrio
				if Aumenta_Altura then
					-- Si el factor de equilibrio es cero, se cambia
					if A.Equilibrio = 0 then
						A.Equilibrio := Factor_Equi;
					else
						-- Si se ha insertado en el subrbol de mayor altura
						-- hay que rotar
						if A.Equilibrio = Factor_Equi then
							-- La rotacin se elige comparando los factores de equilibrio
							-- del nodo crtico (A) y el nodo discriminante (La raz de la
							-- rama donde se ha insertado)
							if Factor_Equi_Hijo = Factor_Equi then
								Rotar_Simple_AVL(A, Rama);
							else
								Rotar_Doble_AVL(A, Rama);
							end if;
						else
							-- Modificacin del factor de equilibrio del nodo crtico
							A.Equilibrio := 0;
						end if;
						-- Se ha restablecido la condicin de equilibrio
						Aumenta_Altura := False;
					end if;
				end if;
			end if;
		end if;
	end Insertar;

	procedure Extraer(
			A: in out rbol_AVL;
			E: in TElemento;
			disminuye_Altura: in out Boolean
			) is

		Procedure Corregir_Factor(
	   			A: in out Rbol_AVL;
				rama: in Integer;
				disminuye_Altura: in out Boolean
				) is
			factor_Eq: Integer range -1 .. 1;
			hijo     : Rbol_AVL := A.Hijos(3 - rama);
		begin
			-- El rbol se "desequilibra" haca el lado contrario
			-- de donde se ha extrado
			if rama = 1 then
				factor_Eq := 1;
			else
				factor_Eq := -1;
			end if;

			if A.Equilibrio = 0 then
				-- Se ha restablecido la condicin de equilibrio
				A.Equilibrio     := -factor_Eq;
				disminuye_Altura := false;
			else	
				if a.equilibrio = factor_Eq then
					-- La extraccin "reequilibra" este subrbol, que
					-- pierde altura
		            A.Equilibrio := 0;
				else
					-- Hay que rotar para restablecer el factor de equilibrio
					-- a los mrgenes permitidos
					if hijo.Equilibrio = factor_Eq then
						rotar_Doble_AVL(A, 3 - rama);
					else
						if hijo.Equilibrio = 0 then
							-- Se va a aplicar una rotacin simple que
						 	-- restablece la condicin de equilibrio
							disminuye_Altura := false;
						end if;
						rotar_Simple_AVL(A, 3 - rama);
					end if;
				end if;
			end if;
		end corregir_Factor;

		procedure Extraer_Sucesor(
				Nodo_Extraer: in out rbol_AVL;
				Nodo_Actual : in out rbol_AVL;
				Disminuye_Altura: in out Boolean
				) is
			Hijo: rbol_AVL;
		begin
			if Nodo_Actual.Hijos(1) = null then
				-- Nodo_Actual es el sucesor
				Hijo               := Nodo_Actual.Hijos(2);
				Nodo_Extraer.Clave := Nodo_Actual.Clave;
				Eliminar(Nodo_Actual);
				Nodo_Actual      := Hijo;
				Disminuye_Altura := True;
			else
				-- El sucesor es descendiente de Nodo_Actual
				Extraer_Sucesor(Nodo_Extraer, Nodo_Actual.Hijos(1), Disminuye_Altura);
				-- Un cambio de altura require ajustar el factor de
			 	-- equilibrio
				if Disminuye_Altura then
					Corregir_Factor(Nodo_Actual, 1, Disminuye_Altura);
				end if;
			end if;
		end Extraer_Sucesor;
		
		Rama: Integer;
		Hijo: rbol_AVL; 
	begin
		if A /= null then
			-- Si E /= A.Clave, la extraccin se producir en
			-- el subrbol que corresponda
 	    	if E < A.Clave then
				Extraer(A.Hijos(1), E, Disminuye_Altura);
				Rama := 1;
			else
				if A.Clave < E then
					Extraer(A.Hijos(2), E, Disminuye_Altura);
					Rama := 2;
				else
					-- Se ha encontrado el nodo a extraer
					-- Tiene grado 0 0 1?
					if A.Hijos(1) = null or else A.Hijos(2) = null then
						if A.Hijos(1) = null then
							Rama := 2;
						else
							Rama := 1;
						end if;
						Hijo := A.Hijos(Rama);
						Eliminar(A);
						A := Hijo;
						Disminuye_Altura := True;
						return;
					else
						-- Tiene grado 2, hay que extraer su sucesor
						-- en simtrico
						Extraer_Sucesor(A, A.Hijos(2), Disminuye_Altura);
						Rama := 2;
					end if;
				end if;
			end if;

			-- Un cambio de altura requiere ajustar el factor de equilibrio
			if Disminuye_Altura then
				Corregir_Factor(A, Rama, Disminuye_Altura);
			end if;
		end if;
	end Extraer;

	function Buscar(A: rbol_AVL; E: in TElemento) return  rbol_AVL is
	begin
		if A = null then
			-- La clave buscada no est en el rbol
			return null;
		else
			if E = A.Clave then
				-- Se ha encontrado
				return A;
			else
				-- La bsqueda contina en el subrbol adecuado
				if E < A.Clave then
					return Buscar(A.Hijos(1),E);
				else
					return Buscar(A.Hijos(2),E);
				end if;
			end if;
		end if;
	end Buscar;

	procedure Vaciar(A: in out rbol_AVL) is
	begin
		-- Para vaciar el rbol se recurre a un recorrido en postorden
		if A /= null then
			Vaciar(A.Hijos(1));
			Vaciar(A.Hijos(2));
			eliminar(A);
		end if;
	end Vaciar;

	function Tamao(A: in rbol_AVL) return Natural is
	begin
		if A = null then
			-- Tamao del rbol vaco
			return 0;
		else
			-- Tamao del rbol no vaco
			return 1 + Tamao(A.Hijos(1)) + Tamao(A.Hijos(2));
		end if;
	end tamao;
	-------------------------------
	-- Operaciones constructoras --
	-------------------------------	
	procedure Insertar(C: in out AVL; E: in TElemento) is
		Aumenta_Altura: Boolean := False;
	begin
		Insertar(C.Elementos, E, Aumenta_Altura);
	end;

	procedure Extraer(C: in out AVL;E: in TElemento) is
		Disminuye_Altura: Boolean := False;
	begin
		Extraer(C.Elementos, E, Disminuye_Altura);
	end;

	procedure Vaciar(C: in out AVL) is
	begin
		Vaciar(C.Elementos);
	end;

	------------------------------
	-- Operaciones observadoras --
	------------------------------
	function Est(C: in AVL; E: in TElemento) return boolean is
	begin
		Return Buscar(C.Elementos, E) /= null;
	end;

	function Tamao(C: in AVL) return Natural is
	begin
		return Tamao(C.Elementos);
	end;
end Contenedores.AVLs;
