#ifndef INC_AVL
#define INC_AVL
#include "autoajustable.h"
#include "nodoavl.h"

template <class T> 
class Avl:public AutoAjustable<T,NodoAvl<T> > {
      NodoAvl<T> *raiz;
      void rotarSimple(NodoAvl<T> *&,int);
      void rotarDoble(NodoAvl<T> *&,int);
      void insertarInterno(NodoAvl<T>*&,const T&,bool &);
      void extraerInterno(NodoAvl<T>*&,const T&,bool &);
      void extraerSucesor(NodoAvl<T>*,NodoAvl<T>*&,bool &);
      void corregirFactor(NodoAvl<T>*&,int,bool &);
   public:
      Avl(){raiz=NULL;}
      ~Avl();
      void insertar(const T&);
      void extraer(const T&);
      bool encuentra(const T&)const;
      void vaciar();
};

// Funciones miembro de la clase Avl
template <class T>
   Avl<T>::~Avl()
   {
      eliminarInterno(raiz);
   }

template <class T>
   void Avl<T>::rotarSimple(NodoAvl<T> *&nodo,int rama)
   {
      int otraRama=1-rama;
      AutoAjustable<T, NodoAvl<T> >::rotarSimple(nodo,rama);
      NodoAvl<T> *hijo=nodo->enlaces[otraRama];
   	//Se ajustan los factores de equilibrio
      hijo->fe=hijo->fe-nodo->fe;
      nodo->fe=-hijo->fe;
   }

template <class T>
   void Avl<T>::rotarDoble(NodoAvl<T> *&nodo,int rama)
   {
      int otraRama=1-rama;
      AutoAjustable<T, NodoAvl<T> >::rotarDoble(nodo,rama);
      NodoAvl<T> *hijo=nodo->enlaces[rama], *otroHijo=nodo->enlaces[otraRama];
   	//Se ajustan los factores de equilibrio
      hijo->fe=(1-2*rama-nodo->fe)/2;
      otroHijo->fe=(2*rama-nodo->fe-1)/2;
      nodo->fe=0;
   }

template <class T>
   void Avl<T>::insertarInterno(NodoAvl<T>*&nodo,const T& valor,bool &aumentaAltura)
   {
      int rama, factorEq;
      if(nodo==NULL)
      {
      	/*El nuevo valor se inserta como hoja*/
         nodo=new NodoAvl<T>(valor);
         this->numElem++;
         aumentaAltura=true;
      }
      else
         /*Si valor es igual a nodo->clave no se inserta*/
         if(valor!=nodo->clave)
         {
         	/*La insercin continua en el subrbol adecuado*/
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            insertarInterno(nodo->enlaces[rama],valor,aumentaAltura);
         	/*Si la altura aumenta hay que comprobar los factores de equilibrio*/
            if(aumentaAltura)
            {
               if(rama==0)
                  factorEq=1;
               else
                  factorEq=-1;
               /*Cambio del factor de equilibrio de los nodos del camino
                desde el nuevo nodo al crtico */
               if(nodo->fe==0)
                  nodo->fe=factorEq;
               else   /*nodo crtico*/
               {
                  /*Si se ha insertado en el subrbol de mayor altura
                   hay que rotar*/
                  if(nodo->fe==factorEq)
                  	/*La rotacin se elige comparando los factores de equilibrio
                  	  del nodo crtico y el nodo discriminante (raz de la rama
                  	  donde se ha insertado)
                  	 */                 
                     if(nodo->enlaces[rama]->fe==factorEq)
                        rotarSimple(nodo,rama);
                     else
                        rotarDoble(nodo,rama);
                  else
                     /*Modificacin del factor de equilibrio del nodo crtico */
                     nodo->fe=0;
                  /*Se ha restablecido la condicin de equilibrio */
                  aumentaAltura=false;
               }
            }  
         } 
   }

template <class T>
   void Avl<T>::corregirFactor(NodoAvl<T>*&nodo,int rama,bool &disminuyeAltura)
   {
      int factorEq;
      NodoAvl<T> *hijo=nodo->enlaces[1-rama];
   	/*El rbol se desequilibra hacia el lado contrario
   	 de donde se ha extrado */
      if(rama==0)
         factorEq=1;
      else
         factorEq=-1;
      if(nodo->fe==0)
      {
       //Se ha restablecido la condicin de equilibrio
         nodo->fe=-factorEq;
         disminuyeAltura=false;
      }
      else
         if(nodo->fe==factorEq)
         /*La extraccin "reequilibra" este subrbol, que
           pierde altura */ 
            nodo->fe=0;
         else
            /*Hay que rotar para restablecer el factor de equilibrio
            a los mrgenes permitidos*/  
            if(hijo->fe==factorEq)
               rotarDoble(nodo,1-rama);
            else
            {
               if(hijo->fe==0)
                  /*Se va a aplicar una rotacin simple que
                    restablece la condicin de equilibrio*/
                  disminuyeAltura=false;
               rotarSimple(nodo,1-rama);
            }
   }

template <class T>
   void Avl<T>::extraerInterno(NodoAvl<T>*&nodo,const T& valor,bool &disminuyeAltura)
   {
      int rama;
      NodoAvl<T> *hijo;
      if(nodo!=NULL)
      {
         if(valor==nodo->clave)
         /*Se ha encontrado el nodo a extraer
         tiene grado 0 o 1?*/
            if((nodo->enlaces[0]==NULL)||(nodo->enlaces[1]==NULL))
            {
               if(nodo->enlaces[0]==NULL)
                  rama=1;
               else
                  rama=0;
               hijo=nodo->enlaces[rama];
               delete nodo;
               nodo=hijo;
               this->numElem--;
               disminuyeAltura=true;
               return;
            }
            else
            {
               /*Tiene grado 2, hay que extraer su sucesor
               en simtrico*/
               extraerSucesor(nodo,nodo->enlaces[1],disminuyeAltura);
               rama=1;
            }
         else
         {
         	/*Si valor es distinto de nodo.clave
         	la extraccin se producir en el subrbol que corresponda*/
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            extraerInterno(nodo->enlaces[rama],valor,disminuyeAltura);
         }
         //Un cambio de altura requiere ajustar el factor de equilibrio
         if(disminuyeAltura) 
            corregirFactor(nodo,rama,disminuyeAltura);
      }
   }

template <class T>
   void Avl<T>::extraerSucesor(NodoAvl<T>*nodoExtraer,NodoAvl<T>*&nodo,bool &disminuyeAltura)
   {
      NodoAvl<T> *hijo;
      if(nodo->enlaces[0]==NULL)
      {
      	//nodo es el sucesor
         hijo=nodo->enlaces[1];
         nodoExtraer->clave=nodo->clave;
         delete nodo;
         nodo=hijo;
         this->numElem--;
         disminuyeAltura=true;
      }
      else
      {
         //El sucesor es descendiente de nodo
         extraerSucesor(nodoExtraer,nodo->enlaces[0],disminuyeAltura);
         //Un cambio de altura require ajustar el factor de equilibrio
         if(disminuyeAltura)
            corregirFactor(nodo,0,disminuyeAltura);
      }
   }

template <class T>
   bool Avl<T>::encuentra(const T& valor)const
   {
      return buscarInterno(raiz,valor);
   }

template <class T>
   void Avl<T>::insertar(const T& valor)
   {
      bool aumentaAltura=false;
      insertarInterno(raiz,valor,aumentaAltura);
   }

template <class T>
   void Avl<T>::extraer(const T& valor)
   {
      bool disminuyeAltura=false;
      extraerInterno(raiz,valor,disminuyeAltura);
   }

template <class T>
   void Avl<T>::vaciar()
   {
      eliminarInterno(raiz);
      raiz=NULL;
      this->numElem=0;
   }

#endif



