package rbolesBinariosBsqueda;

import contenedores.ContenedorOrdenado;

/**
 * Implementacin de un contenedor con un rbol binario de bsqueda 
 */
public class rbolBinarioBsqueda implements ContenedorOrdenado 
{
	// Estructura nodal del rbol binario
	class NodoBinario 
	{
		Comparable clave;
		NodoBinario[] enlaces;
		public NodoBinario(Comparable info) 
		{
			clave   = info;
			enlaces = new NodoBinario[2];
			enlaces[0] = null;
			enlaces[1] = null;
		}
	}

	protected int numElem;

	private NodoBinario raz;

	public rbolBinarioBsqueda() 
	{
		super();
		raz = null;
		numElem = 0;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#est(java.lang.Comparable)
	 */
	public boolean est(Comparable valor) 
	{
		return est(raz, valor);
	}

	protected boolean est(NodoBinario nodo, Comparable valor) 
	{
		while (nodo != null)
			if (valor.compareTo(nodo.clave) == 0)
				return true;
			else if (valor.compareTo((Comparable) nodo.clave) < 0)
				nodo = nodo.enlaces[0];
			else
				nodo = nodo.enlaces[1];
		return false;
	}

	/* (non-Javadoc)
	 * @see contenedores.Contenedor#esVaco()
	 */
	public boolean esVaco() 
	{
		return tamao() == 0;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#extraer(java.lang.Comparable)
	 */
	public void extraer(Comparable valor) 
	{
		raz = extraer(raz, valor);
	}
	
	/**
	 * Extrae un valor del rbol
	 */
	private NodoBinario extraer(NodoBinario nodo, Comparable valor)
	{
		if (nodo != null) 
		{
			if (valor.compareTo(nodo.clave) == 0)
				// Se encontr. El nodo tiene grado 0  1?
				if ((nodo.enlaces[0] == null) || (nodo.enlaces[1] == null)) 
				{
					numElem--;
					if (nodo.enlaces[0] == null)
						return nodo.enlaces[1];
					else
						return nodo.enlaces[0];
				} 
				else
					// Tiene grado 2, hay que buscar su sucesor en simtrico
					nodo.enlaces[1] = extraerSucesor(nodo, nodo.enlaces[1]);
			else 
			{
				// La extraccin contina en el subrbol adecuado
				if (valor.compareTo((Comparable) nodo.clave) < 0)
					nodo.enlaces[0] = extraer(nodo.enlaces[0], valor);
				else
					nodo.enlaces[1] = extraer(nodo.enlaces[1], valor);
			}
		}
		return nodo;
	}

	private NodoBinario extraerSucesor(
		NodoBinario nodoExtraer,
		NodoBinario nodo) 
	{
		if (nodo.enlaces[0] == null) 
		{
			nodoExtraer.clave = nodo.clave;
			numElem--;
			nodo = nodo.enlaces[1];
		} 
		else
			nodo.enlaces[0] = extraerSucesor(nodoExtraer, nodo.enlaces[0]);
		return nodo;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#insertar(java.lang.Comparable)
	 */
	public void insertar(Comparable valor) 
	{
		raz = insertar(raz, valor);
	}

	/**
	 * Inserta un valor en el rbol
	 */
	private NodoBinario insertar(
		NodoBinario nodo, Comparable valor) 
	{
		if (nodo == null) 
		{
			// El nuevo valor se inserta como hoja
			numElem++;
			nodo = new NodoBinario(valor);
		} 
		else 
		{
			// Si valor es igual a nodo.clave no se inserta
			if (valor.compareTo(nodo.clave) != 0) 
			{
				// La insercin contina en el subrbol adecuado
				if (valor.compareTo((Comparable) nodo.clave) < 0)
					nodo.enlaces[0] = insertar(nodo.enlaces[0], valor);
				else
					nodo.enlaces[1] = insertar(nodo.enlaces[1], valor);
			}
		}
		return nodo;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#tamao()
	 */
	public int tamao() 
	{
		return numElem;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#vaciar()
	 */
	public void vaciar() 
	{
		raz = null;
		numElem = 0;
	}
}
